#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import humps
from hsml import predictor_state_condition


class TestPredictorStateCondition:
    # from response json

    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_predictor_state"][
            "response"
        ]["condition"]
        json_camelized = humps.camelize(json)  # as returned by the backend

        # Act
        psc = predictor_state_condition.PredictorStateCondition.from_response_json(
            json_camelized
        )

        # Assert
        assert isinstance(psc, predictor_state_condition.PredictorStateCondition)
        assert psc.type == json["type"]
        assert psc.status == json["status"]
        assert psc.reason == json["reason"]

    # constructor

    def test_constructor(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_predictor_state"][
            "response"
        ]["condition"]

        # Act
        psc = predictor_state_condition.PredictorStateCondition(
            type=json["type"], status=json["status"], reason=json["reason"]
        )

        # Assert
        assert isinstance(psc, predictor_state_condition.PredictorStateCondition)
        assert psc.type == json["type"]
        assert psc.status == json["status"]
        assert psc.reason == json["reason"]

    # extract fields from json

    def test_extract_fields_from_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_predictor_state"][
            "response"
        ]["condition"]

        # Act
        kwargs = (
            predictor_state_condition.PredictorStateCondition.extract_fields_from_json(
                copy.deepcopy(json)
            )
        )

        # Assert
        assert kwargs["type"] == json["type"]
        assert kwargs["status"] == json["status"]
        assert kwargs["reason"] == json["reason"]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import datetime

import humps
from hsml import deployable_component_logs


class TestDeployableComponentLogs:
    # from response json

    def test_from_response_json(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_component_logs_single"][
            "response"
        ]
        json_camelized = humps.camelize(json)
        mocker_from_json = mocker.patch(
            "hsml.deployable_component_logs.DeployableComponentLogs.from_json",
            return_value=None,
        )

        # Act
        dc_logs = deployable_component_logs.DeployableComponentLogs.from_response_json(
            json_camelized
        )

        # Assert
        assert isinstance(dc_logs, list)
        assert len(dc_logs) == 1
        mocker_from_json.assert_called_once_with(json[0])

    def test_from_response_json_list(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_component_logs_list"][
            "response"
        ]
        json_camelized = humps.camelize(json)
        mocker_from_json = mocker.patch(
            "hsml.deployable_component_logs.DeployableComponentLogs.from_json",
            return_value=None,
        )

        # Act
        dc_logs = deployable_component_logs.DeployableComponentLogs.from_response_json(
            json_camelized
        )

        # Assert
        assert isinstance(dc_logs, list)
        assert len(dc_logs) == len(json_camelized)
        assert mocker_from_json.call_count == len(json_camelized)

    def test_from_response_json_empty(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_component_logs_empty"][
            "response"
        ]
        json_camelized = humps.camelize(json)
        mocker_from_json = mocker.patch(
            "hsml.deployable_component_logs.DeployableComponentLogs.from_json",
            return_value=None,
        )

        # Act
        dc_logs = deployable_component_logs.DeployableComponentLogs.from_response_json(
            json_camelized
        )

        # Assert
        assert isinstance(dc_logs, list)
        assert len(dc_logs) == 0
        mocker_from_json.assert_not_called()

    # constructor

    def test_constructor(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["predictor"]["get_deployment_component_logs_single"][
            "response"
        ]
        instance_name = json[0]["instance_name"]
        content = json[0]["content"]
        now = datetime.datetime.now()

        # Act
        dcl = deployable_component_logs.DeployableComponentLogs(
            instance_name=instance_name, content=content
        )

        # Assert
        assert dcl.instance_name == instance_name
        assert dcl.content == content
        assert (dcl.created_at >= now) and (
            dcl.created_at < (now + datetime.timedelta(seconds=1))
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Set, Union

from hopsworks_common.util import (
    FEATURE_STORE_NAME_SUFFIX,
    VALID_EMBEDDING_TYPE,
    Encoder,
    FeatureGroupWarning,
    JobWarning,
    StatisticsWarning,
    StorageWarning,
    ValidationWarning,
    VersionWarning,
    _loading_animation,
    append_feature_store_suffix,
    autofix_feature_name,
    check_timestamp_format_from_date_string,
    convert_event_time_to_timestamp,
    convert_git_status_to_files,
    convert_to_abs,
    feature_group_name,
    generate_fully_qualified_feature_name,
    get_dataset_type,
    get_delta_datestr_from_timestamp,
    get_feature_group_url,
    get_hostname_replaced_url,
    get_hudi_datestr_from_timestamp,
    get_job_url,
    get_timestamp_from_date_string,
    is_interactive,
    is_runtime_notebook,
    run_with_loading_animation,
    strip_feature_store_suffix,
    validate_embedding_feature_type,
    validate_job_conf,
    verify_attribute_key_names,
)
from hsfs import feature, serving_key
from hsfs.core import feature_group_api


if TYPE_CHECKING:
    from hsfs.constructor import serving_prepared_statement


FeatureStoreEncoder = Encoder


def validate_feature(
    ft: Union[str, feature.Feature, Dict[str, Any]],
) -> feature.Feature:
    if isinstance(ft, feature.Feature):
        return ft
    elif isinstance(ft, str):
        return feature.Feature(ft)
    elif isinstance(ft, dict):
        return feature.Feature(**ft)


def parse_features(
    feature_names: Union[
        str, feature.Feature, List[Union[Dict[str, Any], str, feature.Feature]]
    ],
) -> List[feature.Feature]:
    if isinstance(feature_names, (str, feature.Feature)):
        return [validate_feature(feature_names)]
    elif isinstance(feature_names, list) and len(feature_names) > 0:
        return [validate_feature(feat) for feat in feature_names]
    else:
        return []


def build_serving_keys_from_prepared_statements(
    prepared_statements: List[serving_prepared_statement.ServingPreparedStatement],
    feature_store_id: int,
    ignore_prefix: bool = False,
) -> Set[serving_key.ServingKey]:
    serving_keys = set()
    fg_api = feature_group_api.FeatureGroupApi()
    for statement in prepared_statements:
        fg = fg_api.get_by_id(feature_store_id, statement.feature_group_id)
        for param in statement.prepared_statement_parameters:
            serving_keys.add(
                serving_key.ServingKey(
                    feature_name=param.name,
                    join_index=statement.prepared_statement_index,
                    prefix=statement.prefix,
                    ignore_prefix=ignore_prefix,
                    feature_group=fg,
                )
            )
    return serving_keys


__all__ = [
    "FEATURE_STORE_NAME_SUFFIX",
    "VALID_EMBEDDING_TYPE",
    "Encoder",
    "FeatureStoreEncoder",
    "FeatureGroupWarning",
    "JobWarning",
    "StatisticsWarning",
    "StorageWarning",
    "ValidationWarning",
    "VersionWarning",
    "_loading_animation",
    "append_feature_store_suffix",
    "autofix_feature_name",
    "check_timestamp_format_from_date_string",
    "convert_event_time_to_timestamp",
    "convert_git_status_to_files",
    "convert_to_abs",
    "feature_group_name",
    "generate_fully_qualified_feature_name",
    "get_dataset_type",
    "get_delta_datestr_from_timestamp",
    "get_feature_group_url",
    "get_hostname_replaced_url",
    "get_hudi_datestr_from_timestamp",
    "get_job_url",
    "get_timestamp_from_date_string",
    "is_interactive",
    "is_runtime_notebook",
    "run_with_loading_animation",
    "strip_feature_store_suffix",
    "validate_embedding_feature_type",
    "validate_job_conf",
    "verify_attribute_key_names",
    "validate_feature",
    "parse_features",
    "build_serving_keys_from_prepared_statements",
]



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from abc import ABC, abstractmethod
from typing import Optional, Union

import humps
from hopsworks_common import client, util
from hopsworks_common.constants import RESOURCES, Default


class Resources:
    """Resource configuration for a predictor or transformer.

    # Arguments
        cores: Number of CPUs.
        memory: Memory (MB) resources.
        gpus: Number of GPUs.
    # Returns
        `Resources`. Resource configuration for a predictor or transformer.
    """

    def __init__(
        self,
        cores: int,
        memory: int,
        gpus: int,
        **kwargs,
    ):
        self._cores = cores
        self._memory = memory
        self._gpus = gpus

    def describe(self):
        """Print a description of the resource configuration"""
        util.pretty_print(self)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        return Resources(**cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["cores"] = util.extract_field_from_json(json_decamelized, "cores")
        kwargs["memory"] = util.extract_field_from_json(json_decamelized, "memory")
        kwargs["gpus"] = util.extract_field_from_json(json_decamelized, "gpus")
        return kwargs

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {"cores": self._cores, "memory": self._memory, "gpus": self._gpus}

    @property
    def cores(self):
        """Number of CPUs to be allocated per instance"""
        return self._cores

    @cores.setter
    def cores(self, cores: int):
        self._cores = cores

    @property
    def memory(self):
        """Memory resources to be allocated per instance"""
        return self._memory

    @memory.setter
    def memory(self, memory: int):
        self._memory = memory

    @property
    def gpus(self):
        """Number of GPUs to be allocated per instance"""
        return self._gpus

    @gpus.setter
    def gpus(self, gpus: int):
        self._gpus = gpus

    def __repr__(self):
        return f"Resources(cores: {self._cores!r}, memory: {self._memory!r}, gpus: {self._gpus!r})"


class ComponentResources(ABC):
    """Resource configuration for a predictor or transformer.

    # Arguments
        num_instances: Number of instances.
        requests: Minimum resources to allocate for a deployment
        limits: Maximum resources to allocate for a deployment
    # Returns
        `ComponentResource`. Resource configuration for a predictor or transformer.
    """

    def __init__(
        self,
        num_instances: int,
        requests: Optional[Union[Resources, dict, Default]] = None,
        limits: Optional[Union[Resources, dict, Default]] = None,
    ):
        self._num_instances = num_instances
        self._requests = util.get_obj_from_json(requests, Resources) or Resources(
            RESOURCES.MIN_CORES, RESOURCES.MIN_MEMORY, RESOURCES.MIN_GPUS
        )
        self._fill_missing_resources(
            self._requests,
            RESOURCES.MIN_CORES,
            RESOURCES.MIN_MEMORY,
            RESOURCES.MIN_GPUS,
        )
        self._limits = util.get_obj_from_json(limits, Resources) or Resources(
            *self._get_default_resource_limits()
        )
        self._fill_missing_resources(self._limits, *self._get_default_resource_limits())

        # validate both requests and limits
        self._validate_resources(self._requests, self._limits)

    def describe(self):
        """Print a description of the resource configuration"""
        util.pretty_print(self)

    def _get_default_resource_limits(self):
        max_resources = client.get_serving_resource_limits()
        # cores limit
        if max_resources["cores"] == -1:  # no limit
            max_cores = (
                RESOURCES.MAX_CORES
                if RESOURCES.MAX_CORES >= self._requests.cores
                else self._requests.cores
            )
        else:
            max_cores = (
                RESOURCES.MAX_CORES
                if RESOURCES.MAX_CORES <= max_resources["cores"]
                and RESOURCES.MAX_CORES >= self._requests.cores
                else max_resources["cores"]
            )
        # memory limit
        if max_resources["memory"] == -1:  # no limit
            max_memory = (
                RESOURCES.MAX_MEMORY
                if RESOURCES.MAX_MEMORY >= self._requests.memory
                else self._requests.memory
            )
        else:
            max_memory = (
                RESOURCES.MAX_MEMORY
                if RESOURCES.MAX_MEMORY <= max_resources["memory"]
                and RESOURCES.MAX_MEMORY >= self._requests.memory
                else max_resources["memory"]
            )
        # gpus limit
        if max_resources["gpus"] == -1:  # no limit
            max_gpus = (
                RESOURCES.MAX_GPUS
                if RESOURCES.MAX_GPUS >= self._requests.gpus
                else self._requests.gpus
            )
        else:
            max_gpus = (
                RESOURCES.MAX_GPUS
                if RESOURCES.MAX_GPUS <= max_resources["gpus"]
                and RESOURCES.MAX_GPUS >= self._requests.gpus
                else max_resources["gpus"]
            )
        return max_cores, max_memory, max_gpus

    @classmethod
    def _fill_missing_resources(cls, resources, cores, memory, gpus):
        if resources.cores is None:
            resources.cores = cores
        if resources.memory is None:
            resources.memory = memory
        if resources.gpus is None:
            resources.gpus = gpus

    @classmethod
    def _validate_resources(cls, requests, limits):
        # limits
        max_resources = client.get_serving_resource_limits()
        if max_resources["cores"] > -1:
            if limits.cores <= 0:
                raise ValueError("Limit number of cores must be greater than 0 cores.")
            if limits.cores > max_resources["cores"]:
                raise ValueError(
                    "Limit number of cores cannot exceed the maximum of "
                    + str(max_resources["cores"])
                    + " cores."
                )
        if max_resources["memory"] > -1:
            if limits.memory <= 0:
                raise ValueError("Limit memory resources must be greater than 0 MB.")
            if limits.memory > max_resources["memory"]:
                raise ValueError(
                    "Limit memory resources cannot exceed the maximum of "
                    + str(max_resources["memory"])
                    + " MB."
                )
        if max_resources["gpus"] > -1:
            if limits.gpus < 0:
                raise ValueError(
                    "Limit number of gpus must be greater than or equal to 0 gpus."
                )
            if limits.gpus > max_resources["gpus"]:
                raise ValueError(
                    "Limit number of gpus cannot exceed the maximum of "
                    + str(max_resources["gpus"])
                    + " gpus."
                )

        # requests
        if requests.cores <= 0:
            raise ValueError("Requested number of cores must be greater than 0 cores.")
        if limits.cores > -1 and requests.cores > limits.cores:
            raise ValueError(
                "Requested number of cores cannot exceed the limit of "
                + str(limits.cores)
                + " cores."
            )
        if requests.memory <= 0:
            raise ValueError("Requested memory resources must be greater than 0 MB.")
        if limits.memory > -1 and requests.memory > limits.memory:
            raise ValueError(
                "Requested memory resources cannot exceed the limit of "
                + str(limits.memory)
                + " MB."
            )
        if requests.gpus < 0:
            raise ValueError(
                "Requested number of gpus must be greater than or equal to 0 gpus."
            )
        if limits.gpus > -1 and requests.gpus > limits.gpus:
            raise ValueError(
                "Requested number of gpus cannot exceed the limit of "
                + str(limits.gpus)
                + " gpus."
            )

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    @classmethod
    @abstractmethod
    def from_json(cls, json_decamelized):
        pass

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}

        # extract resources
        if cls.RESOURCES_CONFIG_KEY in json_decamelized:
            resources = json_decamelized.pop(cls.RESOURCES_CONFIG_KEY)
        elif "resources" in json_decamelized:
            resources = json_decamelized.pop("resources")
        else:
            resources = json_decamelized

        # extract resource fields
        kwargs["requests"] = util.extract_field_from_json(
            resources, "requests", as_instance_of=Resources
        )
        kwargs["limits"] = util.extract_field_from_json(
            resources, "limits", as_instance_of=Resources
        )

        # extract num instances
        if cls.NUM_INSTANCES_KEY in json_decamelized:
            kwargs["num_instances"] = json_decamelized.pop(cls.NUM_INSTANCES_KEY)
        elif "num_instances" in json_decamelized:
            kwargs["num_instances"] = json_decamelized.pop("num_instances")
        else:
            kwargs["num_instances"] = util.extract_field_from_json(
                resources, [cls.NUM_INSTANCES_KEY, "num_instances"]
            )

        return kwargs

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @abstractmethod
    def to_dict(self):
        pass

    @property
    def num_instances(self):
        """Number of instances"""
        return self._num_instances

    @num_instances.setter
    def num_instances(self, num_instances: int):
        self._num_instances = num_instances

    @property
    def requests(self):
        """Minimum resources to allocate"""
        return self._requests

    @requests.setter
    def requests(self, requests: Resources):
        self._resources = requests

    @property
    def limits(self):
        """Maximum resources to allocate"""
        return self._limits

    @limits.setter
    def limits(self, limits: Resources):
        self._limits = limits

    def __repr__(self):
        return f"ComponentResources(num_instances: {self._num_instances!r}, requests: {self._requests is not None!r}, limits: {self._limits is not None!r})"


class PredictorResources(ComponentResources):
    RESOURCES_CONFIG_KEY = "predictor_resources"
    NUM_INSTANCES_KEY = "requested_instances"

    def __init__(
        self,
        num_instances: int,
        requests: Optional[Union[Resources, dict]] = None,
        limits: Optional[Union[Resources, dict]] = None,
    ):
        super().__init__(num_instances, requests, limits)

    @classmethod
    def from_json(cls, json_decamelized):
        return PredictorResources(**cls.extract_fields_from_json(json_decamelized))

    def to_dict(self):
        return {
            humps.camelize(self.NUM_INSTANCES_KEY): self._num_instances,
            humps.camelize(self.RESOURCES_CONFIG_KEY): {
                "requests": (
                    self._requests.to_dict() if self._requests is not None else None
                ),
                "limits": self._limits.to_dict() if self._limits is not None else None,
            },
        }


class TransformerResources(ComponentResources):
    RESOURCES_CONFIG_KEY = "transformer_resources"
    NUM_INSTANCES_KEY = "requested_transformer_instances"

    def __init__(
        self,
        num_instances: int,
        requests: Optional[Union[Resources, dict]] = None,
        limits: Optional[Union[Resources, dict]] = None,
    ):
        super().__init__(num_instances, requests, limits)

    @classmethod
    def from_json(cls, json_decamelized):
        return TransformerResources(**cls.extract_fields_from_json(json_decamelized))

    def to_dict(self):
        return {
            humps.camelize(self.NUM_INSTANCES_KEY): self._num_instances,
            humps.camelize(self.RESOURCES_CONFIG_KEY): {
                "requests": (
                    self._requests.to_dict() if self._requests is not None else None
                ),
                "limits": self._limits.to_dict() if self._limits is not None else None,
            },
        }



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict

import humps
from hopsworks_common import util


class OnlineIngestionResult:
    """
    Metadata object used to provide Online Ingestion Batch Result information for a feature group.
    """

    def __init__(
        self,
        online_ingestion_id: int = None,
        status: str = None,
        rows: int = None,
        **kwargs,
    ):
        self._online_ingestion_id = online_ingestion_id
        self._status = status
        self._rows = rows

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> OnlineIngestionResult:
        if json_dict is None:
            return None

        json_decamelized: dict = humps.decamelize(json_dict)

        if "count" not in json_decamelized:
            return cls(**json_decamelized)
        elif json_decamelized["count"] == 1:
            return cls(**json_decamelized["items"][0])
        elif json_decamelized["count"] > 1:
            return [cls(**item) for item in json_decamelized["items"]]
        else:
            return None

    def to_dict(self):
        return {
            "onlineIngestionId": self._online_ingestion_id,
            "status": self._status,
            "rows": self._rows,
        }

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @property
    def online_ingestion_id(self) -> int:
        return self._online_ingestion_id

    @property
    def status(self) -> str:
        return self._status

    @property
    def rows(self) -> int:
        return self._rows



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union


if TYPE_CHECKING:
    import great_expectations

import humps
from hopsworks_common.client.exceptions import FeatureStoreException
from hsfs import util
from hsfs.core import expectation_suite_engine
from hsfs.core.constants import (
    HAS_GREAT_EXPECTATIONS,
    initialise_expectation_suite_for_single_expectation_api_message,
)
from hsfs.core.expectation_engine import ExpectationEngine
from hsfs.core.variable_api import VariableApi

# if great_expectations is not installed, we will default to using native Hopsworks class as return values
from hsfs.decorators import uses_great_expectations
from hsfs.ge_expectation import GeExpectation


if HAS_GREAT_EXPECTATIONS:
    import great_expectations


class ExpectationSuite:
    """Metadata object representing a feature validation expectation in the Feature Store."""

    def __init__(
        self,
        expectation_suite_name: str,
        expectations: List[
            Union[
                great_expectations.core.ExpectationConfiguration,
                Dict[str, Any],
                GeExpectation,
            ]
        ],
        meta: Dict[str, Any],
        id: Optional[int] = None,
        run_validation: bool = True,
        validation_ingestion_policy: Literal["always", " strict"] = "always",
        feature_store_id: Optional[int] = None,
        feature_group_id: Optional[int] = None,
        href: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._id = id
        self._expectation_suite_name = expectation_suite_name
        self._ge_cloud_id = kwargs.get("ge_cloud_id", None)
        self._data_asset_type = kwargs.get("data_asset_type", None)
        self._run_validation = run_validation
        self._validation_ingestion_policy = validation_ingestion_policy.upper()
        self._expectations = []
        self._href = href

        if feature_store_id is not None and feature_group_id is not None:
            self._feature_store_id = feature_store_id
            self._feature_group_id = feature_group_id
        elif href is not None:
            self._init_feature_store_and_feature_group_ids_from_href(href)
        else:
            self._feature_group_id = None
            self._feature_store_id = None

        self._variable_api: VariableApi = VariableApi()

        # use setters because these need to be transformed from stringified json
        self.expectations = expectations
        self.meta = meta

        self._expectation_engine: Optional[ExpectationEngine] = None
        self._expectation_suite_engine: Optional[
            expectation_suite_engine.ExpectationSuiteEngine
        ] = None

        if self.id:
            assert (
                self._feature_store_id is not None
            ), "feature_store_id should not be None if expectation suite id is provided"
            assert (
                self._feature_group_id is not None
            ), "feature_group_id should not be None if expectation suite id is provided"
            self._expectation_engine = ExpectationEngine(
                feature_store_id=self._feature_store_id,
                feature_group_id=self._feature_group_id,
                expectation_suite_id=self.id,
            )
            self._expectation_suite_engine = (
                expectation_suite_engine.ExpectationSuiteEngine(
                    feature_store_id=self._feature_store_id,
                    feature_group_id=self._feature_group_id,
                )
            )

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> Optional[Union[ExpectationSuite, List[ExpectationSuite]]]:
        json_decamelized = humps.decamelize(json_dict)
        if (
            "count" in json_decamelized
        ):  # todo count is expected also when providing dict
            if json_decamelized["count"] == 0:
                return None
            return [
                cls(**expectation_suite)
                for expectation_suite in json_decamelized["items"]
            ]
        else:
            return cls(**json_decamelized)

    @classmethod
    @uses_great_expectations
    def from_ge_type(
        cls,
        ge_expectation_suite: great_expectations.core.ExpectationSuite,
        run_validation: bool = True,
        validation_ingestion_policy: Literal["ALWAYS", "STRICT"] = "ALWAYS",
        id: Optional[int] = None,
        feature_store_id: Optional[int] = None,
        feature_group_id: Optional[int] = None,
    ) -> ExpectationSuite:
        """Used to create a Hopsworks Expectation Suite instance from a great_expectations instance.

        # Arguments
            ge_expectation_suite: great_expectations.core.ExpectationSuite
                The great_expectations ExpectationSuite instance to convert to a Hopsworks ExpectationSuite.
            run_validation: bool
                Whether to run validation on inserts when the expectation suite is attached.
            validation_ingestion_policy: str
                The validation ingestion policy to use when the expectation suite is attached. Defaults to "ALWAYS".
                Options are "STRICT" or "ALWAYS".
            id: int
                The id of the expectation suite in Hopsworks. If not provided, a new expectation suite will be created.
            feature_store_id: int
                The id of the feature store of the feature group to which the expectation suite belongs.
            feature_group_id: int
                The id of the feature group to which the expectation suite belongs.

        # Returns
            Hopsworks Expectation Suite instance.
        """
        suite_dict = ge_expectation_suite.to_json_dict()
        if id is None and "id" in suite_dict:
            id = suite_dict.pop("id")
        return cls(
            **suite_dict,
            id=id,
            run_validation=run_validation,
            validation_ingestion_policy=validation_ingestion_policy,
            feature_group_id=feature_group_id,
            feature_store_id=feature_store_id,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self._id,
            "featureStoreId": self._feature_store_id,
            "featureGroupId": self._feature_group_id,
            "expectationSuiteName": self._expectation_suite_name,
            "expectations": self._expectations,
            "meta": json.dumps(self._meta),
            "geCloudId": self._ge_cloud_id,
            "dataAssetType": self._data_asset_type,
            "runValidation": self._run_validation,
            "validationIngestionPolicy": self._validation_ingestion_policy.upper(),
        }

    def to_json_dict(self, decamelize: bool = False) -> Dict[str, Any]:
        the_dict = {
            "id": self._id,
            "featureStoreId": self._feature_store_id,
            "featureGroupId": self._feature_group_id,
            "expectationSuiteName": self._expectation_suite_name,
            "expectations": [
                expectation.to_json_dict() for expectation in self._expectations
            ],
            "meta": self._meta,
            "geCloudId": self._ge_cloud_id,
            "dataAssetType": self._data_asset_type,
            "runValidation": self._run_validation,
            "validationIngestionPolicy": self._validation_ingestion_policy.upper(),
        }

        if decamelize:
            return humps.decamelize(the_dict)
        else:
            return the_dict

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    @uses_great_expectations
    def to_ge_type(self) -> great_expectations.core.ExpectationSuite:
        return great_expectations.core.ExpectationSuite(
            expectation_suite_name=self._expectation_suite_name,
            ge_cloud_id=self._ge_cloud_id,
            data_asset_type=self._data_asset_type,
            expectations=[
                expectation.to_ge_type() for expectation in self._expectations
            ],
            meta=self._meta,
        )

    def _init_feature_store_and_feature_group_ids_from_href(self, href: str) -> None:
        feature_store_id, feature_group_id = re.search(
            r"\/featurestores\/(\d+)\/featuregroups\/(\d+)\/expectationsuite*",
            href,
        ).groups(0)
        self._feature_store_id = int(feature_store_id)
        self._feature_group_id = int(feature_group_id)

    def _init_expectation_engine(
        self, feature_store_id: int, feature_group_id: int
    ) -> None:
        if self.id:
            self._expectation_engine = ExpectationEngine(
                feature_store_id=feature_store_id,
                feature_group_id=feature_group_id,
                expectation_suite_id=self.id,
            )
        else:
            raise ValueError(
                initialise_expectation_suite_for_single_expectation_api_message
            )

    def _init_expectation_suite_engine(
        self, feature_store_id: Optional[int], feature_group_id: Optional[int]
    ) -> None:
        if feature_group_id and feature_store_id:
            self._expectation_suite_engine = (
                expectation_suite_engine.ExpectationSuiteEngine(
                    feature_store_id=feature_store_id,
                    feature_group_id=feature_group_id,
                )
            )
        elif self._feature_group_id and self._feature_store_id:
            self._expectation_suite_engine = (
                expectation_suite_engine.ExpectationSuiteEngine(
                    feature_store_id=self._feature_store_id,
                    feature_group_id=self._feature_group_id,
                )
            )
        else:
            raise ValueError(
                "Provide feature_store_id or feature_group_id to use the expectation suite API"
            )

    # Emulate GE single expectation api to edit list of expectations
    def _convert_expectation(
        self,
        expectation: Union[
            GeExpectation,
            great_expectations.core.ExpectationConfiguration,
            Dict[str, Any],
        ],
    ) -> GeExpectation:
        """
        Convert different representation of expectation to Hopsworks GeExpectation type.

        # Arguments
            expectation: An expectation to convert to Hopsworks GeExpectation type

        # Returns
            An expectation converted to Hopsworks GeExpectation type

        # Raises
            `TypeError`
        """
        if HAS_GREAT_EXPECTATIONS and isinstance(
            expectation, great_expectations.core.ExpectationConfiguration
        ):
            return GeExpectation(**expectation.to_json_dict())
        elif isinstance(expectation, GeExpectation):
            return expectation
        elif isinstance(expectation, dict):
            return GeExpectation(**expectation)
        else:
            raise TypeError(
                "Expectation of type {} is not supported.".format(type(expectation))
            )

    def get_expectation(
        self, expectation_id: int, ge_type: bool = HAS_GREAT_EXPECTATIONS
    ) -> Union[GeExpectation, great_expectations.core.ExpectationConfiguration]:
        """
        Fetch expectation with expectation_id from the backend.

        !!! example
            ```python
            # connect to the Feature Store
            fs = ...

            # get the Feature Group instance
            fg = fs.get_or_create_feature_group(...)

            expectation_suite = fg.get_expectation_suite()
            selected_expectation = expectation_suite.get_expectation(expectation_id=123)
            ```

        # Arguments
            expectation_id: Id of the expectation to fetch from the backend.
            ge_type: Whether to return native Great Expectations object or Hopsworks abstraction,
                defaults to True if great_expectations is installed else false.

        # Returns
            The expectation with expectation_id registered in the backend.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
            `hopsworks.client.exceptions.FeatureStoreException`: If the expectation suite is not registered yet
        """
        if self.id and self._expectation_engine:
            if ge_type:
                return self._expectation_engine.get(expectation_id).to_ge_type()
            else:
                return self._expectation_engine.get(expectation_id)
        else:
            raise FeatureStoreException(
                initialise_expectation_suite_for_single_expectation_api_message
            )

    def add_expectation(
        self,
        expectation: Union[
            GeExpectation, great_expectations.core.ExpectationConfiguration
        ],
        ge_type: bool = HAS_GREAT_EXPECTATIONS,
    ) -> Union[GeExpectation, great_expectations.core.ExpectationConfiguration]:
        """
        Append an expectation to the local suite or in the backend if attached to a Feature Group.

        !!! example
            ```python
            # check if the minimum value of specific column is within a range of 0 and 1
            expectation_suite.add_expectation(
                ge.core.ExpectationConfiguration(
                    expectation_type="expect_column_min_to_be_between",
                    kwargs={
                        "column": "foo_id",
                        "min_value": 0,
                        "max_value": 1
                    }
                )
            )

            # check if the length of specific column value is within a range of 3 and 10
            expectation_suite.add_expectation(
                ge.core.ExpectationConfiguration(
                    expectation_type="expect_column_value_lengths_to_be_between",
                    kwargs={
                        "column": "bar_name",
                        "min_value": 3,
                        "max_value": 10
                    }
                )
            )
            ```
        # Arguments
            expectation: The new expectation object.
            ge_type: Whether to return native Great Expectations object or Hopsworks abstraction,
                defaults to True if great_expectations is installed else false.

        # Returns
            The new expectation attached to the Feature Group.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
            `hopsworks.client.exceptions.FeatureStoreException`: If the expectation suite is not registered yet
        """
        if self.id:
            converted_expectation = self._convert_expectation(expectation=expectation)
            converted_expectation = self._expectation_engine.create(
                expectation=converted_expectation
            )
            self.expectations = self._expectation_engine.get_expectations_by_suite_id()
            if ge_type:
                return converted_expectation.to_ge_type()
            else:
                return converted_expectation
        else:
            raise FeatureStoreException(
                initialise_expectation_suite_for_single_expectation_api_message
            )

    def replace_expectation(
        self,
        expectation: Union[
            GeExpectation, great_expectations.core.ExpectationConfiguration
        ],
        ge_type: bool = HAS_GREAT_EXPECTATIONS,
    ) -> Union[GeExpectation, great_expectations.core.ExpectationConfiguration]:
        """
        Update an expectation from the suite locally or from the backend if attached to a Feature Group.

        !!! example
            ```python
            updated_expectation = expectation_suite.replace_expectation(new_expectation_object)
            ```

        # Arguments
            expectation: The updated expectation object. The meta field should contain an expectationId field.
            ge_type: Whether to return native Great Expectations object or Hopsworks abstraction,
                defaults to True if great_expectations is installed else false.

        # Returns
            The updated expectation attached to the Feature Group.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
            `hopsworks.client.exceptions.FeatureStoreException`: If the expectation suite is not registered yet
        """
        if self.id:
            converted_expectation = self._convert_expectation(expectation=expectation)
            # To update an expectation we need an id either from meta field or from self.id
            self._expectation_engine.check_for_id(converted_expectation)
            converted_expectation = self._expectation_engine.update(
                expectation=converted_expectation
            )
            # Fetch the expectations from backend to avoid sync issues
            self.expectations = self._expectation_engine.get_expectations_by_suite_id()

            if ge_type:
                return converted_expectation.to_ge_type()
            else:
                return converted_expectation
        else:
            raise FeatureStoreException(
                initialise_expectation_suite_for_single_expectation_api_message
            )

    def remove_expectation(self, expectation_id: Optional[int] = None) -> None:
        """
        Remove an expectation from the suite locally and from the backend if attached to a Feature Group.

        !!! example
            ```python
            expectation_suite.remove_expectation(expectation_id=123)
            ```

        # Arguments
            expectation_id: Id of the expectation to remove. The expectation will be deleted both locally and from the backend.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
            `hopsworks.client.exceptions.FeatureStoreException`: If the expectation suite is not registered yet
        """
        if self.id:
            self._expectation_engine.delete(expectation_id=expectation_id)
            self.expectations = self._expectation_engine.get_expectations_by_suite_id()
        else:
            raise FeatureStoreException(
                initialise_expectation_suite_for_single_expectation_api_message
            )

    # End of single expectation API

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        es = "ExpectationSuite("

        if hasattr(self, "id"):
            es += f"id={self._id}, "

        es += f"expectation_suite_name='{self._expectation_suite_name}', "
        es += f"expectations={self._expectations}, "
        es += f"meta={self._meta}, "
        es += f"data_asset_type='{self._data_asset_type}', "
        es += f"ge_cloud_id={self._ge_cloud_id}, "
        es += f"run_validation={self._run_validation}, "
        es += f"validation_ingestion_policy='{self._validation_ingestion_policy}'"

        if hasattr(self, "_feature_group_id"):
            es += f", feature_group_id={self._feature_group_id}, "
            es += f"feature_store_id={self._feature_store_id}"

        es += ")"
        return es

    @property
    def id(self) -> Optional[int]:
        """Id of the expectation suite, set by backend."""
        return self._id

    @id.setter
    def id(self, id: int) -> None:
        self._id = id

    @property
    def expectation_suite_name(self) -> str:
        """Name of the expectation suite."""
        return self._expectation_suite_name

    @expectation_suite_name.setter
    def expectation_suite_name(self, expectation_suite_name: str) -> None:
        self._expectation_suite_name = expectation_suite_name
        if self.id:
            self._expectation_suite_engine.update_metadata_from_fields(
                **humps.decamelize(self.to_dict())
            )

    @property
    def data_asset_type(self) -> Optional[str]:
        """Data asset type of the expectation suite, not used by backend."""
        return self._data_asset_type

    @data_asset_type.setter
    def data_asset_type(self, data_asset_type: Optional[str]) -> None:
        self._data_asset_type = data_asset_type

    @property
    def ge_cloud_id(self) -> Optional[int]:
        """ge_cloud_id of the expectation suite, not used by backend."""
        return self._ge_cloud_id

    @ge_cloud_id.setter
    def ge_coud_id(self, ge_cloud_id: Optional[int]) -> None:
        self._ge_cloud_id = ge_cloud_id

    @property
    def run_validation(self) -> bool:
        """Boolean to determine whether or not the expectation suite shoudl run on ingestion."""
        return self._run_validation

    @run_validation.setter
    def run_validation(self, run_validation: bool) -> None:
        self._run_validation = run_validation
        if self.id:
            self._expectation_suite_engine.update_metadata_from_fields(
                **humps.decamelize(self.to_dict())
            )

    @property
    def validation_ingestion_policy(self) -> Literal["always", "strict"]:
        """Whether to ingest a df based on the validation result.

        "strict" : ingest df only if all expectations succeed,
        "always" : always ingest df, even if one or more expectations fail
        """
        return self._validation_ingestion_policy

    @validation_ingestion_policy.setter
    def validation_ingestion_policy(
        self, validation_ingestion_policy: Literal["always", "strict"]
    ) -> None:
        self._validation_ingestion_policy = validation_ingestion_policy.upper()
        if self.id:
            self._expectation_suite_engine.update_metadata_from_fields(
                **humps.decamelize(self.to_dict())
            )

    @property
    def expectations(self) -> List[GeExpectation]:
        """List of expectations to run at validation."""
        return self._expectations

    @expectations.setter
    def expectations(
        self,
        expectations: Union[
            List[great_expectations.core.ExpectationConfiguration],
            List[GeExpectation],
            List[dict],
            None,
        ],
    ) -> None:
        if expectations is None:
            self._expectations = []
        elif isinstance(expectations, list):
            self._expectations = [
                self._convert_expectation(expectation) for expectation in expectations
            ]
        else:
            raise TypeError(
                f"Type {type(expectations)} not supported. Expectations must be None or a list."
            )

    @property
    def meta(self) -> Dict[str, Any]:
        """Meta field of the expectation suite to store additional information."""
        return self._meta

    @meta.setter
    def meta(self, meta: Union[str, dict]) -> None:
        if isinstance(meta, dict):
            self._meta = meta
        elif isinstance(meta, str):
            self._meta = json.loads(meta)
        else:
            raise ValueError("Meta field must be stringified json or dict.")

        if self._id and hasattr(self, "_expectation_suite_engine"):
            # Adding test on suite_engine allows to not run it on init
            self._expectation_suite_engine.update_metadata_from_fields(
                **humps.decamelize(self.to_dict())
            )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import socket

from hsml import client
from hsml.client.exceptions import ModelRegistryException
from hsml.constants import INFERENCE_ENDPOINTS
from hsml.core import dataset_api, serving_api
from hsml.inference_endpoint import get_endpoint_by_type
from hsml.model_serving import ModelServing


class ModelServingApi:
    def __init__(self):
        self._dataset_api = dataset_api.DatasetApi()
        self._serving_api = serving_api.ServingApi()

    def get(self):
        """Get model serving for specific project.
        :param project: project of the model registry
        :type project: str
        :return: the model serving metadata
        :rtype: ModelServing
        """

        _client = client.get_instance()

        # Validate that there is a Models dataset in the connected project
        if not self._dataset_api.path_exists("Models"):
            raise ModelRegistryException(
                "No Models dataset exists in project {}, Please enable the Serving service or create the dataset manually.".format(
                    _client._project_name
                )
            )

        return ModelServing(_client._project_name, _client._project_id)

    def load_default_configuration(self):
        """Load default configuration and set istio client for model serving"""

        # kserve installed
        is_kserve_installed = self._serving_api.is_kserve_installed()
        client.set_kserve_installed(is_kserve_installed)

        # istio client
        self._istio_init_if_available()

        # resource limits
        max_resources = self._serving_api.get_resource_limits()
        client.set_serving_resource_limits(max_resources)

        # num instances limits
        num_instances_range = self._serving_api.get_num_instances_limits()
        client.set_serving_num_instances_limits(num_instances_range)

        # Knative domain
        knative_domain = self._serving_api.get_knative_domain()
        client.set_knative_domain(knative_domain)

    def _istio_init_if_available(self):
        """Initialize istio client if available"""

        if client.is_kserve_installed():
            # check existing istio client
            try:
                if client.istio.get_instance() is not None:
                    return  # istio client already set
            except Exception:
                pass

            # setup istio client
            inference_endpoints = self._serving_api.get_inference_endpoints()
            if not client._is_external():
                # if internal, get node port
                endpoint = get_endpoint_by_type(
                    inference_endpoints, INFERENCE_ENDPOINTS.ENDPOINT_TYPE_NODE
                )
                if endpoint is not None:
                    client.istio.init(
                        endpoint.get_any_host(),
                        endpoint.get_port(INFERENCE_ENDPOINTS.PORT_NAME_HTTP).number,
                    )
                else:
                    raise ValueError(
                        "Istio ingress endpoint of type '"
                        + INFERENCE_ENDPOINTS.ENDPOINT_TYPE_NODE
                        + "' not found"
                    )
            else:  # if external
                endpoint = get_endpoint_by_type(
                    inference_endpoints, INFERENCE_ENDPOINTS.ENDPOINT_TYPE_LOAD_BALANCER
                )
                if endpoint is not None:
                    # if load balancer (external ip) available
                    _client = client.get_instance()
                    client.istio.init(
                        endpoint.get_any_host(),
                        endpoint.get_port(INFERENCE_ENDPOINTS.PORT_NAME_HTTP).number,
                        _client._project_name,
                        _client._auth._token,  # reuse hopsworks client token
                    )
                    return
                # in case there's not load balancer, check if node port is open
                endpoint = get_endpoint_by_type(
                    inference_endpoints, INFERENCE_ENDPOINTS.ENDPOINT_TYPE_NODE
                )
                if endpoint is not None:
                    # if node port available
                    _client = client.get_instance()
                    host = _client.host
                    port = endpoint.get_port(INFERENCE_ENDPOINTS.PORT_NAME_HTTP).number
                    if self._is_host_port_open(host, port):
                        # and it is open
                        client.istio.init(
                            host,
                            port,
                            _client._project_name,
                            _client._auth._token,  # reuse hopsworks client token
                        )
                        return
                # otherwise, fallback to hopsworks client
                print(
                    "External IP not configured for the Istio ingress gateway, the Hopsworks client will be used for model inference instead"
                )

    def _is_host_port_open(self, host, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)
        try:
            result = sock.connect_ex((host, port))
        finally:
            sock.settimeout(None)
            sock.close()
        return result == 0



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import os
import warnings


# Setting polars skip cpu flag to suppress CPU false positive warning messages printed while importing hsfs
os.environ["POLARS_SKIP_CPU_CHECK"] = "1"

from hsfs import (  # noqa: E402,  Module level import not at top of file because os.environ must be set before importing hsfs
    usage,
    util,
    version,
)
from hsfs.connection import (  # noqa: E402,  Module level import not at top of file because os.environ must be set before importing hsfs
    Connection,
)


__version__ = version.__version__

connection = Connection.connection


def fs_formatwarning(message, category, filename, lineno, line=None):
    return "{}: {}\n".format(category.__name__, message)


warnings.formatwarning = fs_formatwarning
warnings.simplefilter("always", util.VersionWarning)
warnings.filterwarnings(
    action="ignore", category=DeprecationWarning, module=r".*ipykernel"
)


def disable_usage_logging():
    usage.disable()


def get_sdk_info():
    return usage.get_env()


__all__ = ["connection", "disable_usage_logging", "get_sdk_info"]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest
from hsml.decorators import (
    HopsworksConnectionError,
    NoHopsworksConnectionError,
    connected,
    not_connected,
)


class TestDecorators:
    # test not connected

    def test_not_connected_valid(self, mocker):
        # Arrange
        mock_instance = mocker.MagicMock()
        mock_instance._connected = False

        @not_connected
        def assert_not_connected(inst, arg, key_arg):
            assert not inst._connected
            assert arg == "arg"
            assert key_arg == "key_arg"

        # Act
        assert_not_connected(mock_instance, "arg", key_arg="key_arg")

    def test_not_connected_invalid(self, mocker):
        # Arrange
        mock_instance = mocker.MagicMock()
        mock_instance._connected = True

        @not_connected
        def assert_not_connected(inst, arg, key_arg):
            pass

        # Act
        with pytest.raises(HopsworksConnectionError):
            assert_not_connected(mock_instance, "arg", key_arg="key_arg")

    # test connected

    def test_connected_valid(self, mocker):
        # Arrange
        mock_instance = mocker.MagicMock()
        mock_instance._connected = True

        @connected
        def assert_connected(inst, arg, key_arg):
            assert inst._connected
            assert arg == "arg"
            assert key_arg == "key_arg"

        # Act
        assert_connected(mock_instance, "arg", key_arg="key_arg")

    def test_connected_invalid(self, mocker):
        # Arrange
        mock_instance = mocker.MagicMock()
        mock_instance._connected = False

        @connected
        def assert_connected(inst, arg, key_arg):
            pass

        # Act
        with pytest.raises(NoHopsworksConnectionError):
            assert_connected(mock_instance, "arg", key_arg="key_arg")



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import math

import pandas as pd
from hsfs.hopsworks_udf import udf
from hsfs.transformation_statistics import TransformationStatistics


feature_statistics = TransformationStatistics("feature")


@udf(float, drop=["feature"])
def min_max_scaler(feature: pd.Series, statistics=feature_statistics) -> pd.Series:
    return (feature - statistics.feature.min) / (
        statistics.feature.max - statistics.feature.min
    )


@udf(float, drop=["feature"])
def standard_scaler(feature: pd.Series, statistics=feature_statistics) -> pd.Series:
    return (feature - statistics.feature.mean) / statistics.feature.stddev


@udf(float, drop=["feature"])
def robust_scaler(feature: pd.Series, statistics=feature_statistics) -> pd.Series:
    return (feature - statistics.feature.percentiles[49]) / (
        statistics.feature.percentiles[74] - statistics.feature.percentiles[24]
    )


@udf(int, drop=["feature"], mode="pandas")
def label_encoder(feature: pd.Series, statistics=feature_statistics) -> pd.Series:
    unique_data = sorted([value for value in statistics.feature.unique_values])
    value_to_index = {value: index for index, value in enumerate(unique_data)}
    # Unknown categories not present in training dataset are encoded as -1.
    return pd.Series(
        [
            value_to_index.get(data, -1) if not pd.isna(data) else math.nan
            for data in feature
        ]
    )


@udf(bool, drop=["feature"], mode="pandas")
def one_hot_encoder(feature: pd.Series, statistics=feature_statistics) -> pd.Series:
    unique_data = [value for value in statistics.feature.unique_values]

    # One hot encode features. Re-indexing to set missing categories to False and drop categories not in training data statistics.
    # Hence one-hot encoded features would have all categories as False when a category not in training dataset is encountered.
    one_hot = pd.get_dummies(feature, dtype="bool").reindex(
        unique_data, axis=1, fill_value=False
    )

    # Sorting by columns so as to maintain consistency in column order.
    return one_hot.reindex(sorted(one_hot.columns), axis=1)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from hopsworks_common import client
from hopsworks_common.client.exceptions import RestAPIError
from hsfs import decorators, feature_view, training_dataset
from hsfs.constructor import query, serving_prepared_statement
from hsfs.core import explicit_provenance, job, training_dataset_job_conf
from hsfs.core.feature_logging import FeatureLogging
from hsfs.core.job import Job


class FeatureViewApi:
    _POST = "POST"
    _GET = "GET"
    _PUT = "PUT"
    _DELETE = "DELETE"
    _VERSION = "version"
    _QUERY = "query"
    _BATCH = "batch"
    _DATA = "data"
    _PREPARED_STATEMENT = "preparedstatement"
    _TRANSFORMATION = "transformation"
    _TRAINING_DATASET = "trainingdatasets"
    _COMPUTE = "compute"
    _PROVENANCE = "provenance"
    _LINKS = "links"
    _LOGGING = "log"
    _PAUSE_LOGGING = "pause"
    _RESUME_LOGGING = "resume"
    _MATERIALIZE_LOGGING = "materialize"
    _TRANSFORMED_lOG = "transformed"
    _UNTRANSFORMED_LOG = "untransformed"

    def __init__(self, feature_store_id: int) -> None:
        self._feature_store_id = feature_store_id

    @property
    def _client(self):
        return client.get_instance()

    @property
    def _base_path(self):
        return [
            "project",
            self._client._project_id,
            "featurestores",
            self._feature_store_id,
            "featureview",
        ]

    def post(
        self, feature_view_obj: feature_view.FeatureView
    ) -> feature_view.FeatureView:
        headers = {"content-type": "application/json"}
        return feature_view_obj.update_from_response_json(
            self._client._send_request(
                self._POST,
                self._base_path,
                headers=headers,
                data=feature_view_obj.json(),
            )
        )

    def update(self, feature_view_obj: feature_view.FeatureView) -> None:
        headers = {"content-type": "application/json"}
        self._client._send_request(
            self._PUT,
            self._base_path
            + [feature_view_obj.name, self._VERSION, feature_view_obj.version],
            headers=headers,
            data=feature_view_obj.json(),
        )

    @decorators.catch_not_found("hsfs.feature_view.FeatureView", fallback_return=[])
    def get_by_name(self, name: str) -> List[feature_view.FeatureView]:
        """
        Get a feature view from the backend using its name.

        # Arguments
            name `str`: Name of the feature view.

        # Returns
            `List[FeatureView]`: A list that contains all version of the feature view.

        # Raises
            `ValueError`: If the feature group associated with the feature view cannot be found.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        path = self._base_path + [name]
        try:
            return [
                feature_view.FeatureView.from_response_json(fv)
                for fv in self._client._send_request(
                    self._GET,
                    path,
                    {"expand": ["query", "features", "transformationfunctions"]},
                )["items"]
            ]
        except RestAPIError as e:
            if e.response.json().get("errorCode", "") == 270009:
                raise ValueError(
                    "Cannot get back the feature view because the query defined is no longer valid."
                    " Some feature groups used in the query may have been deleted. You can clean up this feature view on the UI"
                    " or `FeatureView.clean`."
                ) from e
            else:
                raise e

    @decorators.catch_not_found("hsfs.feature_view.FeatureView", fallback_return=None)
    def get_by_name_version(self, name: str, version: int) -> feature_view.FeatureView:
        """
        Get a feature view from the backend using both name and version

        # Arguments
            name `str`: Name of feature view.
            version `version`: Version of the feature view.

        # Returns
            `FeatureView`

        # Raises
            `ValueError`: If the feature group associated with the feature view cannot be found.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        path = self._base_path + [name, self._VERSION, version]
        try:
            return feature_view.FeatureView.from_response_json(
                self._client._send_request(
                    self._GET,
                    path,
                    {"expand": ["query", "features", "transformationfunctions"]},
                )
            )
        except RestAPIError as e:
            if e.response.json().get("errorCode", "") == 270009:
                raise ValueError(
                    "Cannot get back the feature view because the query defined is no longer valid."
                    " Some feature groups used in the query may have been deleted. You can clean up this feature view on the UI"
                    " or `FeatureView.clean`."
                ) from e
            else:
                raise e

    def get_all(
        self, latest_version_only: bool = True, with_features: bool = False
    ) -> List[Dict[str, Any]]:
        path = self._base_path
        query_params = {}
        if latest_version_only:
            query_params["filter_by"] = "latest_version"
        if with_features:
            query_params["expand"] = ["features"]
        return self._client._send_request(
            "GET",
            path_params=path,
            query_params={"expand": ["features"]} if with_features else None,
        )["items"]

    def delete_by_name(self, name: str) -> None:
        path = self._base_path + [name]
        self._client._send_request(self._DELETE, path)

    def delete_by_name_version(self, name: str, version: int) -> None:
        path = self._base_path + [name, self._VERSION, version]
        self._client._send_request(self._DELETE, path)

    def get_batch_query(
        self,
        name: str,
        version: int,
        start_time: Optional[int],
        end_time: Optional[int],
        training_dataset_version: Optional[int] = None,
        with_label: bool = False,
        primary_keys: bool = False,
        event_time: bool = False,
        inference_helper_columns: bool = False,
        training_helper_columns: bool = False,
        is_python_engine: bool = False,
    ) -> "query.Query":
        path = self._base_path + [
            name,
            self._VERSION,
            version,
            self._QUERY,
            self._BATCH,
        ]
        return query.Query.from_response_json(
            self._client._send_request(
                self._GET,
                path,
                {
                    "start_time": start_time,
                    "end_time": end_time,
                    "with_label": with_label,
                    "with_primary_keys": primary_keys,
                    "with_event_time": event_time,
                    "inference_helper_columns": inference_helper_columns,
                    "training_helper_columns": training_helper_columns,
                    "is_hive_engine": is_python_engine,
                    "td_version": training_dataset_version,
                },
            )
        )

    def get_serving_prepared_statement(
        self, name: str, version: int, batch: bool, inference_helper_columns: bool
    ) -> List["serving_prepared_statement.ServingPreparedStatement"]:
        path = self._base_path + [
            name,
            self._VERSION,
            version,
            self._PREPARED_STATEMENT,
        ]
        headers = {"content-type": "application/json"}
        query_params = {
            "batch": batch,
            "inference_helper_columns": inference_helper_columns,
        }
        return serving_prepared_statement.ServingPreparedStatement.from_response_json(
            self._client._send_request("GET", path, query_params, headers=headers)
        )

    def create_training_dataset(
        self,
        name: str,
        version: int,
        training_dataset_obj: "training_dataset.TrainingDataset",
    ) -> "training_dataset.TrainingDataset":
        path = self.get_training_data_base_path(name, version)
        headers = {"content-type": "application/json"}
        return training_dataset_obj.update_from_response_json(
            self._client._send_request(
                "POST", path, headers=headers, data=training_dataset_obj.json()
            )
        )

    def get_training_dataset_by_version(
        self, name: str, version: int, training_dataset_version: int
    ) -> "training_dataset.TrainingDataset":
        path = self.get_training_data_base_path(name, version, training_dataset_version)
        return training_dataset.TrainingDataset.from_response_json_single(
            self._client._send_request("GET", path)
        )

    def get_training_datasets(
        self, name: str, version: int
    ) -> List["training_dataset.TrainingDataset"]:
        path = self.get_training_data_base_path(name, version)
        return training_dataset.TrainingDataset.from_response_json(
            self._client._send_request("GET", path)
        )

    def compute_training_dataset(
        self,
        name: str,
        version: int,
        training_dataset_version: int,
        td_app_conf: "training_dataset_job_conf.TrainingDatasetJobConf",
    ) -> job.Job:
        path = self.get_training_data_base_path(
            name, version, training_dataset_version
        ) + [self._COMPUTE]
        headers = {"content-type": "application/json"}
        return job.Job.from_response_json(
            self._client._send_request(
                "POST", path, headers=headers, data=td_app_conf.json()
            )
        )

    def delete_training_data(self, name: str, version: int) -> None:
        path = self.get_training_data_base_path(name, version)
        return self._client._send_request("DELETE", path)

    def delete_training_data_version(
        self, name: str, version: int, training_dataset_version: int
    ) -> None:
        path = self.get_training_data_base_path(name, version, training_dataset_version)
        return self._client._send_request("DELETE", path)

    def delete_training_dataset_only(self, name: str, version: int) -> None:
        path = self.get_training_data_base_path(name, version) + [self._DATA]
        return self._client._send_request("DELETE", path)

    def delete_training_dataset_only_version(
        self, name: str, version: int, training_dataset_version: int
    ) -> None:
        path = self.get_training_data_base_path(
            name, version, training_dataset_version
        ) + [self._DATA]

        return self._client._send_request("DELETE", path)

    def get_training_data_base_path(
        self, name: str, version: int, training_data_version: Optional[int] = None
    ) -> List[Union[str, int]]:
        if training_data_version:
            return self._base_path + [
                name,
                self._VERSION,
                version,
                self._TRAINING_DATASET,
                self._VERSION,
                training_data_version,
            ]
        else:
            return self._base_path + [
                name,
                self._VERSION,
                version,
                self._TRAINING_DATASET,
            ]

    def get_parent_feature_groups(
        self, name: str, version: int
    ) -> explicit_provenance.Links:
        """Get the parents of this feature view, based on explicit provenance.
        Parents are feature groups or external feature groups. These feature
        groups can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature groups, only a minimal information is
        returned.

        # Arguments
            feature_view_instance: Metadata object of feature view.

        # Returns
            `ExplicitProvenance.Links`: the feature groups used to generated this
            feature view
        """
        _client = client.get_instance()
        path_params = self._base_path + [
            name,
            self._VERSION,
            version,
            self._PROVENANCE,
            self._LINKS,
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 1,
            "downstreamLvls": 0,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.UPSTREAM,
            explicit_provenance.Links.Type.FEATURE_GROUP,
        )

    def get_models_provenance(
        self,
        feature_view_name: str,
        feature_view_version: int,
        training_dataset_version: Optional[int] = None,
    ) -> "explicit_provenance.Links":
        """Get the generated models using this feature view, based on explicit
        provenance. These models can be accessible or inaccessible. Explicit
        provenance does not track deleted generated model links, so deleted
        will always be empty.
        For inaccessible models, only a minimal information is returned.

        # Arguments
            feature_view_name: Filter generated models based on feature view name.
            feature_view_version: Filter generated models based on feature view version.
            training_dataset_version: Filter generated models based on the used training dataset version.

        # Returns
            `ExplicitProvenance.Links`: the models generated using this feature
            group
        """
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._PROVENANCE,
            self._LINKS,
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 0,
            "downstreamLvls": 2,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.DOWNSTREAM,
            explicit_provenance.Links.Type.MODEL,
            training_dataset_version=training_dataset_version,
        )

    def enable_feature_logging(
        self,
        feature_view_name: str,
        feature_view_version: int,
    ):
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._LOGGING,
        ]
        _client._send_request("PUT", path_params, {})

    def pause_feature_logging(
        self,
        feature_view_name: str,
        feature_view_version: int,
    ):
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._LOGGING,
            self._PAUSE_LOGGING,
        ]
        return _client._send_request("POST", path_params, {})

    def resume_feature_logging(
        self,
        feature_view_name: str,
        feature_view_version: int,
    ):
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._LOGGING,
            self._RESUME_LOGGING,
        ]
        return _client._send_request("POST", path_params, {})

    def materialize_feature_logging(
        self,
        feature_view_name: str,
        feature_view_version: int,
    ):
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._LOGGING,
            self._MATERIALIZE_LOGGING,
        ]
        jobs_json = _client._send_request("POST", path_params, {})
        jobs = []
        if jobs_json.get("count", 0) > 1:
            for item in jobs_json["items"]:
                jobs.append(Job.from_response_json(item))
        else:
            jobs.append(Job.from_response_json(jobs_json))
        return jobs

    @decorators.catch_not_found(
        "hsfs.core.feature_logging.FeatureLogging", fallback_return=None
    )
    def get_feature_logging(
        self,
        feature_view_name: str,
        feature_view_version: int,
    ):
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._LOGGING,
        ]
        return FeatureLogging.from_response_json(
            _client._send_request("GET", path_params, {})
        )

    def delete_feature_logs(
        self,
        feature_view_name: str,
        feature_view_version: int,
        transformed: bool = None,
    ):
        _client = client.get_instance()
        path_params = self._base_path + [
            feature_view_name,
            self._VERSION,
            feature_view_version,
            self._LOGGING,
        ]
        if transformed is not None:
            if transformed:
                path_params += [self._TRANSFORMED_lOG]
            else:
                path_params += [self._UNTRANSFORMED_LOG]
        _client._send_request("DELETE", path_params, {})



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import functools
import importlib
import os

from hopsworks_common.core.constants import (
    HAS_CONFLUENT_KAFKA,
    HAS_GREAT_EXPECTATIONS,
    HAS_POLARS,
    confluent_kafka_not_installed_message,
    great_expectations_not_installed_message,
    polars_not_installed_message,
)


def not_connected(fn):
    @functools.wraps(fn)
    def if_not_connected(inst, *args, **kwargs):
        if inst._connected:
            raise HopsworksConnectionError
        return fn(inst, *args, **kwargs)

    return if_not_connected


def connected(fn):
    @functools.wraps(fn)
    def if_connected(inst, *args, **kwargs):
        if not inst._connected:
            raise NoHopsworksConnectionError
        return fn(inst, *args, **kwargs)

    return if_connected


def catch_not_found(*class_import_paths, fallback_return=None):
    def decorator(f):
        @functools.wraps(f)
        def g(*args, **kwds):
            # Needs to be imported inside function to avoid circular dependency
            from hopsworks.client.exceptions import RestAPIError
            not_found_error_codes = []
            for class_import_path in class_import_paths:
                class_index = class_import_path.rfind(".")
                module_path = class_import_path[:class_index]
                module = importlib.import_module(module_path)
                object_class = getattr(module, class_import_path[class_index + 1 :])
                not_found_error_codes.append(object_class.NOT_FOUND_ERROR_CODE)
            try:
                return f(*args, **kwds)
            except RestAPIError as e:
                if (
                    e.response.status_code in [400, 404]
                    and e.response.json().get("errorCode", "") in not_found_error_codes
                ):
                    return fallback_return
                else:
                    raise e

        return g

    return decorator


class HopsworksConnectionError(Exception):
    """Thrown when attempted to change connection attributes while connected."""

    def __init__(self):
        super().__init__(
            "Connection is currently in use. Needs to be closed for modification."
        )


class NoHopsworksConnectionError(Exception):
    """Thrown when attempted to perform operation on connection while not connected."""

    def __init__(self):
        super().__init__(
            "Connection is not active. Needs to be connected for feature store operations."
        )


if os.environ.get("HOPSWORKS_RUN_WITH_TYPECHECK", False):
    from typeguard import typechecked
else:
    from typing import TypeVar

    _T = TypeVar("_T")

    def typechecked(
        target: _T,
    ) -> _T:
        return target if target else typechecked


def uses_great_expectations(f):
    @functools.wraps(f)
    def g(*args, **kwds):
        if not HAS_GREAT_EXPECTATIONS:
            raise ModuleNotFoundError(great_expectations_not_installed_message)
        return f(*args, **kwds)

    return g


def uses_polars(f):
    @functools.wraps(f)
    def g(*args, **kwds):
        if not HAS_POLARS:
            raise ModuleNotFoundError(polars_not_installed_message)
        return f(*args, **kwds)

    return g


def uses_confluent_kafka(f):
    @functools.wraps(f)
    def g(*args, **kwds):
        if not HAS_CONFLUENT_KAFKA:
            raise ModuleNotFoundError(confluent_kafka_not_installed_message)
        return f(*args, **kwds)

    return g



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import re
from typing import Optional, Tuple

from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException, RestAPIError


LOADBALANCER_SERVICES = {
    "mysqld": "mysqld",
    "online_store_rest_server": "rdrs",
    "opensearch": "opensearch",
    "kafka": "kafka",
    "feature_query": "flyingduck",
}


class VariableApi:
    def get_variable(self, variable: str):
        """Get the configured value of a variable.

        # Arguments
            variable: Name of the variable.
        # Returns
            The variable's value
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()

        path_params = ["variables", variable]
        domain = _client._send_request("GET", path_params)

        return domain["successMessage"]

    def get_version(self, software: str) -> Optional[str]:
        """Get version of a software component.

        # Arguments
            software: Name of the software.
        # Returns
            The software's version, if the software is available, otherwise `None`.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()

        path_params = ["variables", "versions"]
        resp = _client._send_request("GET", path_params)

        for entry in resp:
            if entry["software"] == software:
                return entry["version"]
        return None

    def parse_major_and_minor(
        self, backend_version: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Extract major and minor version from full version.

        # Arguments
            backend_version: The full version.
        # Returns
            (major, minor): The pair of major and minor parts of the version, or (None, None) if the version format is incorrect.
        """

        version_pattern = r"(\d+)\.(\d+)"
        matches = re.match(version_pattern, backend_version)

        if matches is None:
            return (None, None)
        return matches.group(1), matches.group(2)

    def get_data_science_profile_enabled(self) -> bool:
        """Check if data science profile is enabled on the backend.

        # Returns
            `True`: If data science profile is enabled, `False` otherwise.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self.get_variable("enable_data_science_profile") == "true"

    def get_flyingduck_enabled(self) -> bool:
        """Check if Flying Duck is enabled on the backend.

        # Returns
            `True`: If flying duck is available, `False` otherwise.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self.get_variable("enable_flyingduck") == "true"

    def get_loadbalancer_external_domain(self, service: str) -> str:
        """Get domain loadbalancer for a service.

        # Returns
            `str`: The domain of external loadbalancer for a service, if it is set up.

        # Raises
            `hopsworks.client.exceptions.FeatureStoreException`: If variable is not set in Hopsworks Cluster Configuration.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        try:
            return self.get_variable(f"loadbalancer_external_domain_{service}")
        except RestAPIError as err:
            if err.STATUS_CODE_NOT_FOUND:
                raise FeatureStoreException(
                    f"Client could not get {LOADBALANCER_SERVICES[service]} service hostname from "
                    f"loadbalancer_external_domain_{service}. "
                    "The variable is either not set or empty in Hopsworks cluster configuration."
                ) from err
            else:
                raise err

    def get_service_discovery_domain(self) -> str:
        """Get domain of service discovery server.

        # Returns
            `str`: The domain of service discovery server, if it is set up, otherwise empty string `""`.
        """
        try:
            return self.get_variable("service_discovery_domain")
        except RestAPIError:
            return ""
